unit ScreenDocker;

{*******************************************************

 ScreenDocker
 Copyright 2000 Tan Tze Yong
 
 E-mail: tantzeyong@hotmail.com

 Acknowledgements:
 Portions of the code here is based on the freeware
 TMagneticForm that can be found at www.torry.ru

 ******************************************************}

interface

uses
  Windows, Messages, SysUtils, Classes, Forms, Controls, DsgnIntf;

type
  PRect = ^TRect;

  TScreenEdge = (seLeft, seTop, seRight, seBottom);

  TDockableTo = set of TScreenEdge;

  TDockArea = (daScreen, daWorkArea);

  TOnDockedEvent = procedure (Sender : TObject; var EdgeDocked : TScreenEdge) of object;

  TScreenDockerVersionProperty = class(TStringProperty)
  public
    procedure Edit; override;
    function GetAttributes: TPropertyAttributes; override;
  end;

  TScreenDocker = class(TComponent)
  private
    { Private declarations }
    FAbout          : String;
    FActive         : Boolean;
    FCreating       : Boolean;
    FDockableTo     : TDockableTo;
    FOldWndProc     : TWndMethod;
    FOnDockedEvent  : TOnDockedEvent;
    FOwnerForm      : TForm;
    FDockArea       : TDockArea;
    FDockDistance   : Integer;
    function GetWorkArea : TRect;
    function IsOnlyInstance : Boolean;
    procedure RaiseDockedEvent(EdgeDocked: TScreenEdge);
  protected
    { Protected declarations }
    procedure WndProc(var Msg : TMessage);
    procedure WMMoving(var Msg : TMessage); message WM_MOVING;
  public
    { Public declarations }
    constructor Create(AOwner : TComponent); override;
    destructor Destroy; override;
  published
    { Published declarations }
    property About : String read FAbout write FAbout stored False;
    property Active : Boolean read FActive write FActive default True;
    property DockableTo : TDockableTo read FDockableTo write FDockableTo default [seLeft, seTop, seRight, seBottom];
    property OnDocked :TOnDockedEvent read FOnDockedEvent write FOnDockedEvent;
    property DockArea : TDockArea read FDockArea write FDockArea default daWorkArea;
    property DockDistance : Integer read FDockDistance write FDockDistance default 10;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Samples', [TScreenDocker]);
  RegisterPropertyEditor(Typeinfo(String), TScreenDocker, 'About', TScreenDockerVersionProperty);
end;

{ TScreenDocker }

constructor TScreenDocker.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  // Set flag to true to indicate that our component is middle of being created
  FCreating := True;

  // The owner of the component must be a TForm
  if not (AOwner is TForm) then
    Raise Exception.Create('TScreenDocker can only be placed on a TForm.');

  // We only need one ScreenDocker. Hence we permit one instance of TScreenDocker
  if not IsOnlyInstance then
    Raise Exception.Create('Only one instance of TScreenDocker is allowed per form.');

  FOwnerForm := TForm(AOwner);

  // Replace the form's WndProc with ours so that our component can recieve
  // the form's messages
  FOldWndProc := FOwnerForm.WindowProc;
  FOwnerForm.WindowProc := WndProc;

  // Set the default properties
  FAbout := 'ScreenDocker version 1.00'+#13+#10+'Copyright 2000 Tan Tze Yong'+#13+#10+''+#13+#10+'If you have any bug reports, suggestions and comments, please e-mail '+#13+#10+'me at tantzeyong@hotmail.com';//'ScreenDocker version 1.00';
  FActive := True;
  FDockableTo := [seLeft, seTop, seRight, seBottom];
  FDockArea := daWorkArea;  // Default value set to work area (Screen - Taskbar)
  FDockDistance := 10;      // Default value of 10 pixels

  // Set flag to False as component is created
  FCreating := False;
end;

destructor TScreenDocker.Destroy;
begin
  if not FCreating then
    // Restore the original WndProc
    FOwnerForm.WindowProc := FOldWndProc;

  inherited Destroy;
end;

function TScreenDocker.GetWorkArea: TRect;
begin
  // Retrieve the Rect structure of the work area, which is the screen minus the
  // the taskbar
  SystemParametersInfo(SPI_GETWORKAREA, 0, Addr(Result), 0);
end;

function TScreenDocker.IsOnlyInstance: Boolean;
var
  CompIndex     : Integer;
  CompInstances : Integer;
begin
  CompInstances := 0;

  for CompIndex := 0 to (Owner.ComponentCount - 1) do
    if (Owner.Components[CompIndex] is TScreenDocker) then
      Inc(CompInstances);

  Result := not Boolean(CompInstances > 1);
end;

procedure TScreenDocker.RaiseDockedEvent(EdgeDocked: TScreenEdge);
begin
  if Assigned(OnDocked) then
   OnDocked(Self, EdgeDocked);
end;

procedure TScreenDocker.WMMoving(var Msg: TMessage);
var
  FormDocked   : Boolean;
  DockAreaRect  : TRect;
  DockFormRect  : PRect;
begin
  // Initialise the variable
  FormDocked := False;

  // Check whether form docking is enabled
  if FActive then
  begin
    // Get the RECT structure for the area that our form will be docking to
    case FDockArea of
    daWorkArea:
      DockAreaRect := GetWorkArea;
    daScreen:
      GetWindowRect(GetDesktopWindow, DockAreaRect);
    end;

    // The lParam holds the address of a RECT structure with the
    // screen coordinates of the drag rectangle.
    DockFormRect := Ptr(Msg.LParam);

    // Check whether our form is now within the docking distance of the left
    // edge of the dock area.
    if (seLeft in FDockableTo) then
      if (DockFormRect^.Left <= FDockDistance) and (DockFormRect^.Left >= -(FDockDistance)) then
      begin
        DockFormRect^.Left := 0;
        DockFormRect^.Right := FOwnerForm.Width;
        RaiseDockedEvent(seLeft);
        FormDocked := True;
      end;

    // Check whether our form is now within the docking distance of the top
    // edge of the dock area.
    if (seTop in FDockableTo) then
      if (DockFormRect^.Top <= FDockDistance) and (DockFormRect^.Top >= -(FDockDistance)) then
      begin
        DockFormRect^.Top := 0;
        DockFormRect^.Bottom := FOwnerForm.Height;
        RaiseDockedEvent(seTop);
        FormDocked := True;
      end;

    // Check whether our form is now within the docking distance of the right
    // edge of the dock area.
    if (seRight in FDockableTo) then
      if (DockFormRect^.Right >= (DockAreaRect.Right - FDockDistance)) and (DockFormRect^.Right <= (DockAreaRect.Right + FDockDistance)) then
      begin
        DockFormRect^.Right := DockAreaRect.Right;
        DockFormRect^.Left := (DockAreaRect.Right - FOwnerForm.Width);
        RaiseDockedEvent(seRight);
        FormDocked := True;
      end;

    // Check whether our form is now within the docking distance of the bottom
    // edge of the dock area.
    if (seBottom in FDockableTo) then
      if (DockFormRect^.Bottom >= (DockAreaRect.Bottom - FDockDistance)) and (DockFormRect^.Bottom <= (DockAreaRect.Bottom + FDockDistance)) then
      begin
        DockFormRect^.Bottom := DockAreaRect.Bottom;
        DockFormRect^.Top := (DockAreaRect.Bottom - FOwnerForm.Height);
        RaiseDockedEvent(seBottom);
        FormDocked := True;
      end;
  end;

  // The Win32 SDK recommends that we set the Result to True to inform Windows
  // that we have process the message
  Msg.Result := Integer(FormDocked);

  FOldWndProc(Msg);
end;

procedure TScreenDocker.WndProc(var Msg: TMessage);
begin
  // Check whether the message is the one we want to process
  if Msg.Msg = WM_MOVING then
    WMMOVING(Msg)
  else
    FOldWndProc(Msg);
end;

{ TScreenDockerVersionProperty }

procedure TScreenDockerVersionProperty.Edit;
begin
  MessageBox(Application.Handle, PChar('ScreenDocker version 1.00'+#13+#10+'Copyright 2000 Tan Tze Yong'+#13+#10+''+#13+#10+'If you have any bug reports, suggestions and comments, please e-mail '+#13+#10+'me at tantzeyong@hotmail.com'), 'Information', MB_ICONINFORMATION);
end;

function TScreenDockerVersionProperty.GetAttributes: TPropertyAttributes;
begin
  Result := [paDialog];
end;

end.
